//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: prom_node.C,v 1.20 2004/08/20 07:19:04 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           prom_node.C
// 
#include "prom_grid.hh"

PromTable<PromDTet*> *PromGeoNode::bufftable_ = NULL;

static const int blocksize = 1024;

#undef __FUNCT__  
#define __FUNCT__  "PromParentBank::GetFreePar"
/* PromParentBank::GetFreePar() *************************************
 * 
 *  - private helpers to allocate a par
 *
 */
int PromParentBank::GetFreePar( PromParent** newel )
{
  *newel = NULL;  

  if ( numParFree_ == 0 ) {
    if( freePar_ != NULL ) {
      *newel = freePar_;
      freePar_ = dynamic_cast<PromParent*>(freePar_->next_);
      (*newel)->next_ = NULL;
    }
    else {
      currPar_ = new PromParent[blocksize]; CHKNEW(currPar_,1);
      for(int ii=0;ii<blocksize;ii++)currPar_[ii].kill();
      if( !parBlocks_ ) {
	parBlocks_ = new PromList<PromParent*>(10); CHKNEW(parBlocks_,1);
      }
      parBlocks_->addTail( currPar_ );
      numParFree_ = blocksize;
    }
  }

  // give next if necc.
  if ( *newel == NULL ) {
    *newel = currPar_++; // this is the return value
    numParFree_--;
  }
  (*newel)->kill(FALSE);
  assert((*newel)->next_ == NULL);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromParentBank::Deallocate"
/* PromParentBank::Deallocate() *************************************
 *
 *  - set free internally for reuse
 */
int PromParentBank::Deallocate( PromParent *elem )
{
  assert(parBlocks_ != NULL);

  if ( freePar_ ) elem->next_ = freePar_;
  else assert(elem->next_ == NULL);
  freePar_ = elem;
  elem->SetPsz( 0 );
  if(elem->DAP_ != NULL) PetscFree(elem->DAP_); elem->DAP_ = NULL;
  elem->kill();
  elem->setProc(-9); elem->setGlobal(-9); 

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromParentBank::ClearCache"
/* PromParentBank::ClearCache *********************************************
 *
 *  - returns number of levels currently allocated
 */
int PromParentBank::ClearCache()
{   
  if( parBlocks_ ){
    while( !parBlocks_->isEmpty() ) {
      PromParent *par = parBlocks_->removeHead();  
      delete [] par;   
    }
    delete parBlocks_;  parBlocks_ = NULL;
  } 
  freePar_ = NULL;
  numParFree_ = 0;
  currPar_ = NULL;

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "CPlex::create"
CPlex* CPlex::create( CPlex*& pHead, UINT nMax, UINT cbElement )
{ 
  assert(nMax > 0); assert(cbElement > 0);
  PetscFunctionBegin;
  size_t sz = sizeof(CPlex) + nMax*cbElement;
  CPlex* p; int ierr = PetscMalloc(sz,&p); assert(ierr==0);
  
  p->nMax = nMax;
  p->nCur = 0;
  p->pNext = pHead;
  pHead = p;  // change head (adds in reverse order for simplicity)
  PetscFunctionReturn(p);
}

#undef __FUNCT__  
#define __FUNCT__ "CPlex::freeDataChain"
void CPlex::freeDataChain()     // free this one and links
{ 
  PetscFunctionBegin;
  CPlex* p = this;
  while (p != NULL) {
    CPlex* pN = p->pNext;
    BYTE* bytes = (BYTE*)p; 
    int ierr = PetscFree( bytes ); 
    if(ierr!=0){
      fprintf(stdout,"PetscFree error");
    }
    if(ierr)PetscPrintf(PETSC_COMM_SELF,"[?]%s ierr=%d\n",
			__FUNCT__,ierr);
    p = pN;
  }
  PetscStackPop; /* since function returns void cannot use PetscFunctionReturn(); */
}

//////////////////////////////////////////////////////////////////////
// PromGlobalID
//
PromGlobalID::PromGlobalID( int i /*=-9*/, int p /*=-9*/ )
{
  assert(this!=NULL);    
  lIndex_ = i;
  procID_ = p; 
} 
PromGlobalID::~PromGlobalID()
{
  assert(this!=NULL);
}
PromGlobalID &PromGlobalID::operator=( const PromGlobalID &a )
{
  assert(this!=NULL);
  lIndex_ = a.lIndex_; 
  procID_ = a.procID_;
  return *this;
}
 
//////////////////////////////////////////////////////////////////////
// PromCR
//
#undef __FUNCT__    
#define __FUNCT__ "PromCR::PromCR"
/* PromCR::PromCR() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
PromCR::PromCR(int **ppb) : gnodes_(NULL), vals_(NULL)
{
  assert(this!=NULL); 
  bits_.eq = 0;  bits_.len = 0; bits_.flag = 0; 
  Read( ppb );
}

PromCR::PromCR( int nn ) : gnodes_(NULL), vals_(NULL)
  
{
  bits_.eq = 0;  bits_.len = nn; bits_.flag = 0; 
  assert(this!=NULL); assert(nn>0);
  PetscMalloc(nn*sizeof(int),&gnodes_); 
}

#undef __FUNCT__    
#define __FUNCT__ "PromCR::~PromCR"
/* PromCR::~PromCR() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
PromCR::~PromCR()
{
  PetscFree( gnodes_ );
  if ( vals_ ) PetscFree( vals_ );
}

#undef __FUNCT__  
#define __FUNCT__ "PromCR::GetWrite_size"
/* PromCR::GetWrite_size ***************************************
 *
 */
int PromCR::GetWrite_size( int &sz )const
{
  assert(this!=NULL);
  sz = bits_.len + 2;
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromCR::Read"
/* PromCR::Read ***************************************
 *
 */
int PromCR::Read( int **ppb )
{
  assert(this!=NULL);
  int ierr, *pb = *ppb; unsigned int ii;
  if( bits_.len > 0 ) PetscFree(gnodes_);
  
  // <eq, len, len*<gids> >
  bits_.eq = *pb++; bits_.len = *pb++; assert(bits_.len>0);
  ierr = PetscMalloc( bits_.len*sizeof(int), &gnodes_ ); CHKERRQ(ierr);
  for( ii = 0 ; ii < bits_.len ; ii++ ) gnodes_[ii] = *pb++;
  *ppb = pb;
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromCR::Write"
/* PromCR::Write ***************************************
 *
 */
int PromCR::Write( int **ppb )const
{
  unsigned int ii; int *pb = *ppb; assert(this!=NULL);
  
  // <eq, len, len*<gids> >
  *pb++ = bits_.eq; *pb++ = bits_.len;
  for( ii = 0 ; ii < bits_.len ; ii++ ) *pb++ = gnodes_[ii];
  *ppb = pb;
  
  return 0;
}

//////////////////////////////////////////////////////////////////////
// PromNodeBase
//
PromNodeBase &PromNodeBase::operator=( const PromNodeBase &a )
{
  assert(this != NULL);
  int state = bits_.state;
  PromElement::operator=( a );      bits_.state = state;
  index_ = a.index_;
  geq_ = a. geq_;
  return *this;
}

//////////////////////////////////////////////////////////////////////
// PromNode
//
#undef __FUNCT__  
#define __FUNCT__ "PromNode::~PromNode" 
/* PromNode::~PromNode() ***********************************************
 * 
 */
PromNode::~PromNode()
{
  assert(this!=NULL);
  if( parents_ != NULL ) delete parents_; 
}

#undef __FUNCT__  
#define __FUNCT__ "PromNode::reset" 
/*  PromNode::reset ******************************************
 *
 */
void PromNode::reset()
{
  assert(this!=NULL); 
 
  PromElement::reset();
  bits_.dead = 1;
  
  geq_ = -8;
  index_.reset();
  if( symbData_ ) {
    if( symbData_->dependents ){ 
      while( !symbData_->dependents->isEmpty() ) {
	delete symbData_->dependents->removeHead();
      }
      delete symbData_->dependents; symbData_->dependents = NULL; 
    }
    symbData_->ghostID.reset();
    symbData_->adjac.removeAll();
    symbData_->elemList.removeAll();
    symbData_->child = symbData_->parent = NULL;
  }
}

#undef __FUNCT__  
#define __FUNCT__ "PromSymbData::PromSymbData" 
/* PromSymbData::PromSymbData ******************************************
 *
 */
PromSymbData::PromSymbData() : 
  elemList(10),
  adjac(0), // this must be alllocated 
  child(NULL),
  dependents(NULL),
  parent(NULL) 
{ 
}

#undef __FUNCT__  
#define __FUNCT__ "PromSymbData::~PromSymbData" 
/* PromSymbData::~PromSymbData ******************************************
 *
 */
PromSymbData::~PromSymbData()
{
  assert(this!=NULL);

  if( dependents != NULL ){
    while( !dependents->isEmpty() ) {
      delete dependents->removeHead();
    }
    delete dependents; dependents = NULL;
  }
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::reset" 
/*  PromGeoNode::reset ******************************************
 *
 */
void PromGeoNode::reset()
{
  assert(this!=NULL); PromNode::reset();
  if( symbGeoData_ ) {
    for(int i=0;i<6;i++) symbGeoData_->faceID[i] = -9;
    symbGeoData_->parElem = NULL;
  }
}

#undef __FUNCT__  
#define __FUNCT__ "PromAggNode::reset" 
/*  PromAggNode::reset ******************************************
 *
 */
void PromAggNode::reset()
{
  assert(this!=NULL); PromNode::reset();
}

#undef __FUNCT__  
#define __FUNCT__ "PromNode::LiftCopy" 
/* PromNode::LiftCopy******************************************
 *
 *   INPUT:
 *     - a:symbolic node
 *
 *   SIDE EFFECTS: 
 *   RETURN:
 *     - this
 */
int PromNode::LiftCopy( const PromNode * const fine )
{
  assert(this!=NULL);

  PromElement::operator=( *fine );  assert(bits_.shp_sz==0); assert(bits_.qual==0); 
  assert(bits_.malloced==0); // bits_.qual = 0;  
  index_ = fine->index_;                 assert(nParents()==0);
  assert(bits_.ndf == fine->bits_.ndf);  // NO parents_,...
  for( int i = 0; i < PROM_MAXNDF ; i++ ) setFixed(i,FALSE);
  if( fine->symbData_ != NULL ) {
    symbData_->ghostID = fine->symbData_->ghostID;
  }
  // no geq

  return 0; 
}

/* PromNode::Copy******************************************
 *
 *   INPUT:
 *     - a: MPI node
 *
 *   RETURN:
 *     - this
 */
int PromNode::Copy( const PromMPINode &a )
{
  assert(this!=NULL);
  PromElement::operator=( a );
  index_ = a.index_; assert(bits_.ndf == a.bits_.ndf);                
  if( symbData_ ) symbData_->ghostID = a.ghostID_; 
  geq_ = a.geq();                  assert( geq_>=0);
  
  return 0; 
}

/* PromGeoNode::Copy******************************************
 *
 *   INPUT:
 *     - a: MPI node
 *
 *   RETURN:
 *     - this
 */
int PromGeoNode::Copy( const PromMPINode &a )
{
  assert(this!=NULL);
  int ierr = PromNode::Copy( a ); CHKERRQ(ierr);
  if( symbGeoData_ ) symbGeoData_->coord = a.coord_; 
  
  return 0; 
}

#undef __FUNCT__  
#define __FUNCT__ "PromNode::getParent" 
/* PromNode::getParent******************************************
 *
 *   INPUT:
 *   SIDE EFFECTS:
 *   RETURN:
 */
PromNode* PromNode::getParent() const 
{ 
  assert(this!=NULL);  
  
  PromNode* curr = NULL;  PromNodeState st;       assert(symbData_!=NULL);
  if( (st=getState()) == PROM_SELECTED ) curr = symbData_->parent; 
  else if( st == PROM_DELETED && symbData_->parent ) {
    curr = symbData_->parent->symbData_->parent;
  }
  else assert(st == PROM_DELETED); // needs to be set
 
  return dynamic_cast<PromNode*>(curr); // bug: can be nul
}

#undef __FUNCT__  
#define __FUNCT__ "PromNode::SetnParents" 
/* PromNode::SetnParents ******************************************
 *
 */ 
int PromNode::SetnParents( PromParentBank * const bank, const int nn )
{
  assert(this!=NULL);   assert(nn>=0);
  int ierr;
  if(nn > PROM_MAXPAR_HARD){ 
    SETERRQ2(1,"SetnParents(%d) >= PROM_MAXPAR(%d)",nn,PROM_MAXPAR_HARD);
  }
  if( parents_==NULL && nn ){ 
    parents_ = new PromParentArr( bank, nn ); CHKNEW(parents_,1);
  }
  else if( parents_ && parents_->getSize() != nn ){ // reallocate
    ierr = parents_->SetSize( bank, nn ) ; CHKERRQ(ierr);
  }

  return 0;
}
  
#undef __FUNCT__  
#define __FUNCT__ "PromNode::LiftCopyIn" 
/* PromNode::LiftCopyIn ******************************************
 *
 */  
int PromNode::LiftCopyIn( PromMPINode *mpinode ) const
{
  mpinode->PromNodeBase::operator=( *this );      
  mpinode->fIndex_ = -7;
  assert(symbData_!=NULL);
  if( isGhost() ) mpinode->ghostID_ = symbData_->ghostID;
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromAggNode::LiftCopyIn" 
/* PromAggNode::LiftCopyIn ******************************************
 *
 */  
int PromAggNode::LiftCopyIn( PromMPINode *mpinode ) const
{
  int ierr, ndf;
  ierr = PromNode::LiftCopyIn( mpinode ); CHKERRQ(ierr);  
  ndf = getNextNDF();
  ierr = mpinode->SetNDF( ndf ); CHKERRQ(ierr); 
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromAggNode::getNextNDF" 
/* PromAggNode::getNextNDF ******************************************
 *
 */  
int PromAggNode::getNextNDF() const
{
  int nndf, fndf = getNDF(), rsz = getBsz(); // cndf is encode in 
  if( rsz > 0 ) {
    nndf = rsz/fndf;                             assert(rsz%fndf==0);
  }
  else {
    assert(symbData_ != NULL);
    if(symbData_->parent != NULL) nndf = symbData_->parent->getNDF();
    else nndf = 0;
  }
  return nndf;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::LiftCopyIn"
/* PromGeoNode::LiftCopyIn ******************************************
 *
 */  
int PromGeoNode::LiftCopyIn( PromMPINode *mpinode ) const
{
  int ierr;
  ierr = PromNode::LiftCopyIn(mpinode);   CHKERRQ(ierr);    
  assert(symbGeoData_!=NULL);
  mpinode->coord_ = symbGeoData_->coord;  
  return 0;
}

//////////////////////////////////////////////////////////////////////
// PromAggNode
//
#undef __FUNCT__  
#define __FUNCT__ "PromAggNode::PromAggNode" 
/* PromAggNode::PromAggNode ******************************************
 *
 */  
PromAggNode::PromAggNode() :
  PromNode(),  
  B_(NULL),
  Bsz_(0) 
{ 
}
/* PromAggNode::~PromAggNode ******************************************
 *
 */  
#undef __FUNCT__  
#define __FUNCT__ "PromAggNode::~PromAggNode" 
PromAggNode::~PromAggNode()
{ 
  if( B_ ){ PetscFree(B_); B_ = NULL; Bsz_ = 0; }
}
 
#undef __FUNCT__  
#define __FUNCT__ "PromAggNode::Write" 
/* PromAggNode::Write ******************************************
 * 
 *  INPUT:
 *    - ppb: buffer to write/read into/from
 *
 *  RETURN:
 *    - PETSc error code.
 */
int PromAggNode::Write( int **ppb )const
{
  int ii, ierr, *pb = *ppb;
  
  if((size_t)pb%sizeof(shpfloat) != 0){ SETERRQ(1,"unaligned access"); };
  // global,bsz,npar,geq,bsz*shape,PromElement,npar*<[-3],parent>,[-2]
  *pb++ = index_.global();
  *pb++ = Bsz_;
  *pb++ = nParents();
  *pb++ =  geq_;                                  assert( geq_>=0);
  assert((size_t)pb%sizeof(shpfloat)==0);
  shpfloat ff, *pf = (shpfloat*)pb;
  for( ii = 0 ; ii < Bsz_ ; ii++ ){ ff = B_[ii]; *pf++ = ff; }
  pb = (int*)pf;
  ierr = PromElement::Write( &pb ); CHKERRQ(ierr);
  for( ii = 0 ; ii < nParents() ; ii++ ){
    if((size_t)pb%sizeof(shpfloat)!=0) *pb++ = -3; //makeeven
    ierr = parents_->data_[ii]->Write( &pb ); CHKERRQ(ierr); 
  }
  if((size_t)pb%sizeof(shpfloat)!=0) *pb++ = -2; //makeeven
  *ppb = pb;
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromAggNode::Read" 
/* PromAggNode::Read ******************************************
 *
 */
int PromAggNode::Read( PromParentBank * const bank, int **ppb )
{
  int ii, ierr, *pb = *ppb, nb, np;
  assert(nParents() == 0 && parents_ == NULL);

  // <global,bsz,npar,geq,bsz*shape,PromElement,npar*<[-3],parent>,[-2]
  ii = *pb++; index_.setGlobal( ii );
  nb = *pb++; ierr = SetBsz( nb ); CHKERRQ(ierr);
  np = *pb++;
  geq_ = *pb++; 
 
  assert( (size_t)pb%sizeof(shpfloat) == 0 ); //makeeven
  shpfloat ff, *pf = (shpfloat*)pb;
  for( ii = 0 ; ii < Bsz_ ; ii++ ){ ff = *pf++; B_[ii] = ff; }
  pb = (int*)pf;
  ierr = PromElement::Read( &pb );   CHKERRQ(ierr);
  assert( !parents_ );
  ierr = SetnParents( bank, np );      CHKERRQ(ierr);
  for( ii=0 ; ii < np ; ii++ ){ 
    if((size_t)pb%sizeof(shpfloat)!=0) pb++; //makeeven
    ierr = parents_->data_[ii]->Read( &pb );   CHKERRQ(ierr);
  }
  if((size_t)pb%sizeof(shpfloat)!=0){ ii = *pb++; assert(ii == -2);}
  *ppb = pb;

  return 0;  
}

#undef __FUNCT__  
#define __FUNCT__ "PromAggNode::GetWrite_size" 
/* PromAggNode::GetWrite_size ******************************************
 *
 */
int PromAggNode::GetWrite_size( int &sz )const
{
  int ii,ierr,tt,sisz=sizeof(shpfloat)/sizeof(int); 
  // global,bsz,npar,geq,bsz*shape,PromElement,npar*<[-3],parent>,[-2]
  sz = 4;// gid, npar, bsz,geq
  sz += Bsz_ * sisz; // B array
  ierr = PromElement::GetWrite_size(tt); CHKERRQ(ierr);
  sz += tt; // geq 
  for( ii = 0 ; ii < nParents() ; ii++ ){
    if(sz%sisz != 0) sz++;   
    ierr = parents_->data_[ii]->GetWrite_size(tt); CHKERRQ(ierr);
    sz += tt;
  }
  if(sz%sisz != 0) sz++;   

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromAggNode::SetBsz" 
/* PromAggNode::SetBsz ******************************************
 *
 */  
int PromAggNode::SetBsz(int ss)
{ 
  if(B_ && Bsz_ != ss){ PetscFree(B_); B_ = NULL; }
  if(ss && B_==NULL){int ierr=PetscMalloc(sizeof(shpfloat)*ss, &B_);CHKERRQ(ierr);}
  Bsz_ = ss;  assert( (Bsz_ == 0) == (B_ == NULL));

  return 0;
}

/* PromAggNode::Die ******************************************
 *
 */  
int PromAggNode::Die( PromBank *bank ) 
{ 
  int ierr = bank->Deallocate( this ); return ierr; 
}

#undef __FUNCT__  
#define __FUNCT__ "PromAggNode::LiftCopy" 
/* PromAggNode::LiftCopy******************************************
 *
 *   INPUT:
 *     - a:symbolic node
 *
 *   SIDE EFFECTS: 
 *   RETURN:
 *     - error code
 */
int PromAggNode::LiftCopy( const PromNode * const a )
{
  assert(this!=NULL);   
  const PromAggNode * const fine = dynamic_cast<const PromAggNode*>(a);
  assert(fine!=NULL);

  PromNode::LiftCopy( fine );
  int ierr, ndf;
  ndf = fine->getNextNDF(); 
  ierr = SetNDF(ndf); CHKERRQ(ierr); 
  
  return 0;
}

//////////////////////////////////////////////////////////////////////
// PromGeoNode
//
#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::Die" 
/* PromGeoNode::Die ******************************************
 *
 */  
int PromGeoNode::Die( PromBank *bank )
{ 
  int ierr = bank->Deallocate( this ); return ierr;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::PromGeoNode"
/* PromGeoNode::PromGeoNode ******************************************
 *
 */  
PromGeoNode::PromGeoNode() :
  PromNode()
{
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::~PromGeoNode" 
/* PromGeoNode::~PPromGeoNode ******************************************
 *
 */ 
PromGeoNode::~PromGeoNode()
{ 
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::LiftCopy" 
/* PromGeoNode::LiftCopy******************************************
 *
 *   INPUT:
 *     - a:symbolic node
 *
 *   SIDE EFFECTS: 
 *   RETURN:
 *     - this
 */
int PromGeoNode::LiftCopy( const PromNode * const a )
{
  assert(this!=NULL); 
  const PromGeoNode * const fine = dynamic_cast<const PromGeoNode*>(a);
  
  PromNode::LiftCopy(a);

  if( fine->symbGeoData_ ) {
    assert(symbGeoData_!=NULL);
    symbGeoData_->coord = fine->symbGeoData_->coord;
    symbGeoData_->faceID[0] = fine->symbGeoData_->faceID[0];
    symbGeoData_->faceID[1] = fine->symbGeoData_->faceID[1];
    symbGeoData_->faceID[2] = fine->symbGeoData_->faceID[2];
    symbGeoData_->faceID[3] = fine->symbGeoData_->faceID[3];
    symbGeoData_->faceID[4] = fine->symbGeoData_->faceID[4];
    symbGeoData_->faceID[5] = fine->symbGeoData_->faceID[5];
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::Write" 
/* PromGeoNode::Write ******************************************
 *
 *  INPUT:
 *    - ppb: buffer to write/read into/from
 *
 *  RETURN:
 *    - PETSc error code.
 */
int PromGeoNode::Write(int **ppb)const
{
  int *pb,ierr,ii;
  //coord,global,geq,npar,PromElement,npar*<[-3],parent>, faces*,[-2]
  if( symbGeoData_ ) {
    ierr = symbGeoData_->coord.Write( ppb );   CHKERRQ(ierr); 
  }
  pb = *ppb; 
  *pb++ = index_.global();                        assert(geq_>=0);
  *pb++ = geq_; 
  *pb++ = nParents();  
  ierr = PromElement::Write( &pb ); CHKERRQ(ierr);
  for( ii = 0 ; ii < nParents() ; ii++ ){
    if((size_t)pb%sizeof(shpfloat)!=0) *pb++ = -3; //makeeven
    ierr = parents_->data_[ii]->Write( &pb ); CHKERRQ(ierr); 
  }
  // no adjacs 
  if( symbGeoData_ ){ ierr = WriteFaceIDs( pb );         CHKERRQ(ierr); }
  if((size_t)pb%sizeof(shpfloat)!=0) *pb++ = -2; //makeeven
  *ppb = pb;
  
  return 0;
}
#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::Read" 
/* PromGeoNode::Read ******************************************
 *
 */
int PromGeoNode::Read( PromParentBank * const bank, int **ppb )
{
  int *pb, ierr, np, ii; 
  
  //coord,local,global,geq,npar,PromElement,npar*<[-3],par>,faces*,[-2]
  if( symbGeoData_ ) {
    ierr = symbGeoData_->coord.Read( ppb ); CHKERRQ(ierr); 
  }
  pb = *ppb; 
  ii = *pb++;           index_.setGlobal(ii);
  geq_ = *pb++; 
  np = *pb++; 
  ierr = PromElement::Read( &pb );   CHKERRQ(ierr);
  assert( !parents_ );
  ierr = SetnParents( bank, np );      CHKERRQ(ierr);
  for( ii=0 ; ii < np ; ii++ ){ 
    if((size_t)pb%sizeof(shpfloat)!=0) pb++; //makeeven
    ierr = parents_->data_[ii]->Read( &pb );   CHKERRQ(ierr);
  }
  // no adjacs 
  if( symbGeoData_ ) ReadFaceIDs( pb );
  if((size_t)pb%sizeof(shpfloat)!=0){ ii = *pb++; assert(ii == -2);}
  *ppb = pb;
  
  return 0;
}
#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::GetWrite_size" 
/* PromGeoNode::GetWrite_size ******************************************
 * 
 *  -- no adjacs
 */
int PromGeoNode::GetWrite_size( int &sz )const
{
  int ii,tt,sisz=sizeof(shpfloat)/sizeof(int),ierr; 
  // coord,<local,global>,geq,npar,PromElement,npar*<[-3],parent>,faces*,[-2]
  sz = 3; // gid, npar, geq
  if( symbGeoData_ ){
    ierr = symbGeoData_->coord.GetWrite_size( ii ); sz += ii; CHKERRQ(ierr);
  }
  ierr = PromElement::GetWrite_size(tt); CHKERRQ(ierr);
  sz += tt; 
  for( ii = 0 ; ii < nParents() ; ii++ ){
    if(sz%sisz != 0) sz++;   
    ierr = parents_->data_[ii]->GetWrite_size(tt); CHKERRQ(ierr);
    sz += tt;
  }
  // no adjacs 
  if( symbGeoData_ ) sz += numFaces() + 1;
  if(sz%sisz != 0) sz++;   
  
  return 0;
}

// PromGeoNode methods

#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::SetFaceIDs" 
/* PromGeoNode::SetFaceIDs******************************************
 *
 *   INPUT:
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromGeoNode::SetFaceIDs( int id0, int id1/*-9*/, int id2/*-9*/, 
			     int id3/*-9*/, int id4/*-9*/, int id5/*-9*/ )
{ 
  assert(this!=NULL);  assert(symbGeoData_!=NULL);
  symbGeoData_->faceID[0] = id0;   
  symbGeoData_->faceID[1] = id1;   
  symbGeoData_->faceID[2] = id2; 
  symbGeoData_->faceID[3] = id3;   
  symbGeoData_->faceID[4] = id4;   
  symbGeoData_->faceID[5] = id5;
  assert(id0 < id1 || id1 == -9); assert(id1 < id2 || id2 == -9);
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::MaxFaceID" 
/* PromGeoNode::MaxFaceID******************************************
 *
 *   INPUT:
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromGeoNode::MaxFaceID()const
{
  assert(this!=NULL); assert(symbGeoData_!=NULL);
  int ii, ret=0;
  for(ii=0;ii<6 && symbGeoData_->faceID[ii]!=-9;ii++) 
    if(symbGeoData_->faceID[ii] > ret) ret = symbGeoData_->faceID[ii];

  return ret;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::MaxOff" 
/* PromGeoNode::MaxOff******************************************
 *
 *   INPUT:
 *     - smax: max off shape (in/out)
 *     - maxi: index of smax (out)
 *     - buff: copy this shape if it modifies 'smax'
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromGeoNode::MaxOff(gfloat &smax, int *maxi/*=0*/, gfloat *buff/*=0*/)const
{ 
  int i,j;

  for(i = 0 ; i < nParents() ; i++ ) {
    if( parents_->data_[i]->getPsz() == 0 ) continue;
    if( parents_->data_[i]->getPsz() == 1 ){
      assert(bits_.pointer==0);
      if( -parents_->data_[i]->shape_ > smax || 
	  parents_->data_[i]->shape_ - 1.0 > smax ){
	if(buff)
	  for(j=0;j<nParents();j++) buff[j] = parents_->data_[j]->shape_;
	if( -parents_->data_[i]->shape_ > smax ) 
	  smax = -parents_->data_[i]->shape_;
	else smax = parents_->data_[i]->shape_ - 1.0;
	if( maxi ) *maxi = i;
      }
    }
    else assert(0);
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::setElemWithList" 
/* PromGeoNode::setElemWithList *********************************************
 *
 *  - find a node with a elem list to start
 *
 *  INPUT:
 *     - eList: list of elements to start check
 *     - bank:
 *     - mype:
 *     - numtries: number of iteration to do 
 *     - tol: tolerence to start at.
 *     - set: flag to set node in element
 *     - nested: are the grid nodes nested
 *
 *  SIDE EFFECTS:
 *     - if 'flag' == TRUE then, the found element has node added to its child 
 *     - sets 'node' shape values
 *
 *  RETURN:
 *     - 
 */
PromDTet *PromGeoNode::setElemWithList( const PromList<PromPoly*> *eList,
					PromParentBank * const bank, const int mype,
					int numtries, gfloat tol,
					const bool set, const bool nested)
{
  assert(this!=NULL); assert(tol>=0.0); assert(numtries>0);
  PromDTet *elem = NULL;
  int deep = 20, shallow = 3;
 
  while( numtries-- > 0 ) {
    // use elem list of parent to see
    PromDTet  *elem1 = NULL;
    gfloat fout, minfout=0.0;  PromDTet *alt;
    Position pos = eList->getHeadPosition();              assert(pos!=NULL);
    while( pos ) {
      elem = dynamic_cast<PromDTet*>(eList->getNext(pos));
      assert(!elem->isDead());
      // check if not dead - happens as list are not cleared emediatly
      assert( bufftable_ == NULL || bufftable_->isEmpty() );
      elem = setElem(elem, bank, mype, tol, set, shallow, NULL, nested, 
		     &alt, &fout);
      if ( elem ) return elem;
      if( alt && (!elem1 || fout < minfout) ) { 
	elem1 = alt; minfout = fout; 
      }
    }
    if( elem1 ){
      elem = setElem ( elem1, bank, mype, tol, set, deep, NULL, nested );
      if ( elem ) return elem;
    }
    else assert(0);
    tol *= 10.0; deep += 20; shallow += 3;
  }
  
  return elem;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::setElem" 
/* PromGeoNode::setElem *************************************************
 *
 *  - find the element that a node is in, and set its shape function values
 *  - this is on a finer grid and 'elem' is on the coarser grid also
 *  - NOT reentrant
 *
 *  INPUT:
 *    - elem: an element (on next grid ) to check for this node
 *    - bank:
 *    - mype:
 *    - tol: tolerence to use for 'isIn()'
 *    - set: set parent, and put into parents list
 *    - levels: number of elemets in chain to check (recursive levels)
 *    - nodo: element not to do
 *    - nested: flag for if the grids are still nested
 *    - best: best elem found (out)
 *    - fout: off shape of 'best' (in/out)
 *
 *  SIDE EFFECTS:
 *    - sets nodes shape fuction values
 *    - sets 'parElem' (pointer to element on this grid - this must get fixed
 *    - maintains  'parElem' list of fine nodes
 *    - set 'parElem' to NULL if not found
 *
 *  RETURN:
 *    - element that it is in
 */
PromDTet *PromGeoNode::setElem( PromDTet *elem, PromParentBank * const bank, 
				const int mype, const gfloat tol,
				const bool set, const int levels, 
				const PromDTet* nodo, const bool nested,
				PromDTet **best/*=NULL*/, 
				gfloat *fout/*=NULL*/)
{ 
  assert(this!=NULL); 
  PromDTet       *el;
  shpfloat  tempshape[4] = {0.0, 0.0, 0.0, 0.0};

  // NOT reentrant - pass it arround for multi thread
  if(!bufftable_){
    bufftable_ = new PromTable<PromDTet*>(227); CHKNEW_void(bufftable_,1);
  }
  if( bufftable_->isEmpty() && best ) { *best = NULL; *fout = 1.e10; }
  bufftable_->add( elem ); // keep table of done e;ems

  // look for a child node of elemes nodes
  if ( getState() == PROM_SELECTED && elem != nodo ){ 
    // look for this in my nodes
    for ( int j = 0 ; j < 4 ; j++ ) {
      PromGeoNode *n1 = dynamic_cast<PromGeoNode*>(elem->nodes_[j]);
      if( (n1->symbData_->child == this && nested) || 
	  ((!nested || n1->symbData_->child == NULL) &&
	   (fabs(n1->symbGeoData_->coord[0]-symbGeoData_->coord[0])<1.e-16) &&
	   (fabs(n1->symbGeoData_->coord[1]-symbGeoData_->coord[1])<1.e-16) &&
	   (fabs(n1->symbGeoData_->coord[2]-symbGeoData_->coord[2])<1.e-16))){
	// set it
	if( set ) {
	  symbGeoData_->parElem = elem;    assert(isChildListEmpty());        
	  // set shape
	  int ierr = SetnParents(bank,1); assert(ierr==0);
	  PromParent *par = parents_->data_[0];
	  par->Copy( n1, mype );
	  ierr = parents_->data_[0]->setScalarP0(1.0); assert(ierr==0);
	  elem->addChild ( this );
	}
	bufftable_->removeAll();	  assert(elem!=NULL);
	return elem;
      }
    }
  }  

  // check if is here
  bool ok = TRUE; // !elem->isSliver ()
  if( elem != nodo ) {
    if( elem->isIn( symbGeoData_->coord, tempshape, tol, mype, &ok ) ) {
      if( set ) {
	symbGeoData_->parElem = elem;	assert(isChildListEmpty()); 
	int ierr = SetnParents(bank,4); assert(ierr==0); 
	for( int j = 0 ; j < 4 ; j++ ) {
	  PromNode *n1 = elem->nodes_[j]; PromParent *par = parents_->data_[j];
	  par->Copy( n1, mype );
	  int ierr = parents_->data_[j]->setScalarP0(tempshape[j]);
	  assert(ierr==0);
	} 
	elem->addChild ( this );
      }
      bufftable_->removeAll(); assert(elem!=NULL); 
      return elem;
    }
    else if( ok && best ){
      gfloat maxx = 0.0; // look for an ik one
      for( int i = 0 ; i < 4 ; i++ ) {
	if( tempshape[i] < 0.0 && maxx < -tempshape[i] ) maxx = -tempshape[i];
	else if( maxx < tempshape[i] - 1.0 ) maxx = tempshape[i] - 1.0;
      }
      assert(maxx>0.0);
      if( fout && maxx < *fout ) { *fout = maxx; *best = elem; } 
    }
  }
  // get a candidate
  shpfloat lowshp1 = 1.e10;
  PromDTet *adjel1 = NULL;
  for ( int j = 0; j < 4 ; j++ ) {
    el = dynamic_cast<PromDTet*>(elem->neigh_[j]);
    assert( !el || !el->isDead() );
    
    if ( el && !bufftable_->find(el) && el != nodo ) {
      if ( ok ) { 
        if ( lowshp1 > tempshape[j] ) {
          // look for best direction
          lowshp1 = tempshape[j];
          adjel1 = el;
        }
      }
      else if( el && !adjel1 ) { // !ok - sliver
	PromGeoNode *n1,*n2,*n3;
	n1 = dynamic_cast<PromGeoNode*>(elem->nodes_[PromPoly::tetID_[j][0]]);
	n2 = dynamic_cast<PromGeoNode*>(elem->nodes_[PromPoly::tetID_[j][1]]);
	n3 = dynamic_cast<PromGeoNode*>(elem->nodes_[PromPoly::tetID_[j][2]]);
	if(PromContext::crossDot( n3->symbGeoData_->coord, 
				  n2->symbGeoData_->coord, 
				  n1->symbGeoData_->coord, 
				  symbGeoData_->coord) > 0.0) adjel1 = el;
      }
      else adjel1 = el;
    } 
  }
  // check best candidate
  if ( adjel1 && levels > 0 ) {
    return setElem( adjel1, bank, mype, tol, set, levels - 1, nodo, nested,
		    best, fout);
  }
  else {
    bufftable_->removeAll(); 
    return NULL;   
  }
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::SetElemAdjac"
/* PromGeoNode::SetElemAdjac **************************************
 *
 *   - make 'neigh' of elems WITHIN this continuous set of elems (all tets)
 *   - this is in the center of the elements 
 *
 *   INPUT:
 *   SIDE EFFECTS:
 *   RETURN:
 *     - 
 */
int PromGeoNode::SetElemAdjac()const 
{
  PromNode   *n1, *n2;
  Position    pos1, pos2;
  PromDTet   *elem1,*elem2;
  int         count = 0;                          assert(symbGeoData_!=NULL);
  const int   numels = 3 * symbData_->elemList.getCount() / 2;
  assert( symbData_->elemList.getCount()%2 == 0 ); 
  
  // for all elems (elem1)
  pos1 = symbData_->elemList.getHeadPosition(); 
  while ( pos1 ) {
    elem1 = dynamic_cast<PromDTet*>(symbData_->elemList.getNext(pos1)); 
    if( elem1->neigh_[0] && elem1->neigh_[1] && elem1->neigh_[2] ) continue;
    
    // for all elems (elem2)
    pos2 = symbData_->elemList.getHeadPosition();
    while ( pos2 ) {
      elem2 = dynamic_cast<PromDTet*>(symbData_->elemList.getNext(pos2));      
      if (elem2->neigh_[0] && elem2->neigh_[1] && elem2->neigh_[2]) continue;
      
      // for all interior faces of 'elem1'
      for ( int ii = 0 ; ii < 3 ; ii++ ) {
        if ( elem1->neigh_[ii] ) continue;     // done
        n1 = elem1->nodes_[PromPoly::tetID_[ii][1]];   // test node
        
        // for all interior faces of 'elem2'
        for ( int jj = 0 ; jj < 3 ; jj++ ) {
          if ( elem2->neigh_[jj] ) continue;   // done
          n2 = elem2->nodes_[PromPoly::tetID_[jj][2]]; // test node
          
          // 'neigh_' are in order
          if (n1 == n2 &&
              elem1->nodes_[PromPoly::tetID_[ii][2]] == 
	      elem2->nodes_[PromPoly::tetID_[jj][1]]) {
            elem1->neigh_[ii] = elem2;
            elem2->neigh_[jj] = elem1;
            count++;                        // counts edges on a convex surface
            if ( count == numels ) {
              return count;
            }
          }
        }  
      }
    }    
  }
  
  assert(0);
  return 1;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::MakeExteriorFacets"
/* PromGeoNode::MakeExteriorFacets **************************************
 *
 *   - make faces exterior faces of this 
 *
 *   INPUT:
 *     - facets: array to put new facets in
 *     - edges: arrays to put edges in of each facet 
 *     - totvol: return value of the volume of the elements used
 *     - str:
 *     - topfacet: number of facets (out)
 *
 *   SIDE EFFECTS:
 *     - creates new facets
 *
 *   RETURN:
 *     - 
 */
int PromGeoNode::MakeExteriorFacets( PromFacet *facets[], 
				     PromGeoNode *edges[][2],
				     gfloat *totvol, PromBank *bank, 
				     int &topfacet ) const
{  
  int ii, oposite, nface;
  
  topfacet = 0;
  Position pos = symbData_->elemList.getHeadPosition();
  while ( pos ) {
    PromDTet *elem = dynamic_cast<PromDTet*>(symbData_->elemList.getNext(pos));
    *totvol = *totvol + elem->getVolume();
    
    for( ii = nface = oposite = 0 ; ii < 4 ; ii++ ) {
      if( !elem->neigh_[ii] ) {
	if(elem->nodes_[ii] == dynamic_cast<const PromNode*>(this)) oposite++;
	else {
	  nface++;
	  // reverse to point out
	  PromNode *n1 = elem->nodes_[PromPoly::tetID_[ii][2]]; 
	  PromNode *n2 = elem->nodes_[PromPoly::tetID_[ii][1]]; 
	  PromNode *n3 = elem->nodes_[PromPoly::tetID_[ii][0]];
	  PromFacet  *facet = bank->newFacet( n1, n2, n3 );
	  facets[topfacet] = facet;
	  
	  int found = 0;
	  for ( int jj = 0 ; jj < 3 ; jj++ ) {              
	    if( elem->nodes_[PromPoly::tetID_[ii][jj]] == 
		dynamic_cast<const PromNode*>(this) ){
	      n1 = elem->nodes_[PromPoly::tetID_[ii][(jj+2)%3]];
	      n2 = elem->nodes_[PromPoly::tetID_[ii][(jj+1)%3]];
	      edges[topfacet][0] = dynamic_cast<PromGeoNode*>(n1);
	      edges[topfacet][1] = dynamic_cast<PromGeoNode*>(n2);
	      topfacet++;
	      found++;
	      if ( topfacet == 50 ) {
		*totvol = 0.0; // flag for error
		return 0;
	      }
	      break;
	    }
	  }
	  assert( found );
	}
      }
    }
  }
  assert( topfacet < 50);
 
  return 0;
}

//////////////////////////////////////////////////////////////////////
// PromParent
//
#undef __FUNCT__  
#define __FUNCT__ "PromParent::~PromParent" 
/* PromParent::~PromParent() ***********************************************
 *
 */
PromParent::~PromParent()
{
  if( bits_.malloced ){ 
    assert(P0_ != NULL && bits_.pointer); 
    PetscFree(P0_);
  }
  if( DAP_ ){ PetscFree(DAP_); DAP_ = NULL;}
} 

#undef __FUNCT__  
#define __FUNCT__ "PromParent::GetWrite_size" 
/* PromParent::GetWrite_size ******************************************
 *
 */
int PromParent::GetWrite_size( int &sz )const
{
  int ierr, tt = getPsz(), sisz = sizeof(shpfloat)/sizeof(int); 
  ierr = PromElement::GetWrite_size(sz); CHKERRQ(ierr);  // sendfloat
  sz += 4; // gIndex, procID, shpsz, geq
  
  // shape data 
  if( sz%sisz != 0 ) sz++;  // make even
  if( tt > 0 ) {
    if((bits_.pointer && P0_ != NULL) || bits_.pointer==0) sz += tt * sisz; // P0/shape
    if( DAP_ != NULL ) sz += tt * sisz; // P0/shape
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromParent::Write" 
/* PromParent::Write ******************************************
 *
 *  INPUT:
 *    - ppb: buffer to write/read into/from (in/out)
 *
 *  RETURN: 
 *    - PETSc error code.
 */
int PromParent::Write( int **ppb ) const
{
  int *pb = *ppb, tt; assert(bits_.aux == 0);

  assert((size_t)pb%sizeof(shpfloat)==0);
  struct bitsTAG bits = bits_;  
  bits.aux = (DAP_==NULL) ? 0 : 2;
  bits.aux += ((bits.pointer && P0_ != NULL) || bits.pointer==0) ? 1 : 0;
  struct bitsTAG *pbt = (struct bitsTAG*)pb; *pbt++ = bits; pb = (int*)pbt;
  *pb++ = gIndex_;  
  *pb++ = procID_; 
  *pb++ = geq_;                            assert(geq_>=0);
  *pb++ = tt = getPsz();

  if((size_t)pb%sizeof(shpfloat)!=0) *pb++ = -3; // make even  
  if( tt > 0 ) {
    if( bits.aux >= 2 ) { // code for DAP
      shpfloat *fpb = (shpfloat*)pb;      assert(DAP_ != NULL);
      for( int ii = 0 ; ii < tt ; ii++ ) *fpb++ = DAP_[ii]; 
      pb = (int*)fpb;
    }
    shpfloat *fpb = (shpfloat*)pb;
    if( bits.aux%2 == 1 && bits.pointer ) { // 'aux' code for P0_ 
      if(P0_==NULL){ for( int ii = 0 ; ii < tt ; ii++ ) *fpb++ = 0.0; assert(0); } //???
      else for( int ii = 0 ; ii < tt ; ii++ ) *fpb++ = P0_[ii]; 
    }
    else if( bits.pointer==0 ){ *fpb++ = shape_; assert(tt==1); }
    pb = (int*)fpb;
  }
  //else SETERRQ1(1,"getPsz == 0 %d",tt);
  bits.aux = 0;

  *ppb = pb;

  return 0;
}
#undef __FUNCT__  
#define __FUNCT__ "PromParent::Read" 
/* PromParent::Read ******************************************
 * 
 */ 
int PromParent::Read( int **ppb  )
{
  int ierr,tt,*pb=*ppb,ii; assert(bits_.aux == 0);
  assert((size_t)pb%sizeof(shpfloat)==0); 
  ierr = PromElement::Read( &pb );  CHKERRQ(ierr); 
  gIndex_ = *pb++; 
  procID_ = *pb++;  
  geq_ = *pb++; 
  tt = *pb++; 

  if((size_t)pb%sizeof(shpfloat)!=0) pb++; //makeeven
  if( tt > 0 ) {
    if( bits_.aux >= 2 ) { // code for have DAP_
      assert(DAP_ == NULL); // length is ndf() * fine->ndf()
      ierr  = PetscMalloc( tt*sizeof(shpfloat), &DAP_ );   CHKERRQ(ierr); 
      shpfloat *vv = DAP_; shpfloat *fpb = (shpfloat*)pb; 
      for(ii=0;ii<tt;ii++) *vv++ = *fpb++;
      pb = (int*)fpb; assert(bits_.pointer);
      if( bits_.aux%2 == 0 ) { // !P0 -- only DAP
	SetPsz_hard( tt ); 
	bits_.pointer = 0; P0_ = NULL;
      }
    }
    // P0
    if( bits_.aux%2 == 1 ) {
      shpfloat *fpb = (shpfloat*)pb; 
      if( bits_.pointer ) { // code for P0_
	ierr = SetPsz( tt ); CHKERRQ(ierr);// allocs shapes
	shpfloat *vv = P0_; 
	if(vv==NULL)SETERRQ1(1,"P0_==NULL, sz=%d",tt);
	for(ii=0;ii<tt;ii++) *vv++ = *fpb++;
      }
      else if(bits_.pointer == 0){ shape_ = *fpb++; bits_.shp_sz = 1; }
      pb = (int*)fpb;
    }
  }
  else assert(DAP_==NULL && bits_.pointer); 
  bits_.aux = 0;

  *ppb = pb;

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromParent::SetPsz" 
/* PromParent::SetPsz ******************************************
 *  
 */
int PromParent::SetPsz( int i )
{ 
  assert(this!=NULL); //if(i<=0)SETERRQ(1,"i<=0");;
  if(i < 0 || i > 4095){ SETERRQ1(1,"SetPsz(%d) > 16383",i); }

  PetscFunctionBegin;  
  if( (int)bits_.shp_sz == i ) PetscFunctionReturn(0); // nothing to do
  if( bits_.pointer ) {
    bits_.pointer = 0;       assert(P0_==NULL && !bits_.malloced);
  } 
  bits_.shp_sz = i;
  if( bits_.shp_sz > 1 ){
    int tt = bits_.shp_sz;
    int ierr = PetscMalloc( sizeof(shpfloat)*tt, &P0_);   CHKERRQ(ierr);
    bits_.malloced = 1; bits_.pointer = 1;
    while( tt-- ) P0_[tt] = PROM_BIG_NUM;
  }
  else assert(bits_.malloced == 0 && bits_.pointer == 0);
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "PromParent::setScalarP0" 
int PromParent::setScalarP0( shpfloat val )
{
  assert(this!=NULL);
  if( getPsz() > 1 ) SETERRQ(1,"getPsz() > 1");
  bits_.shp_sz = 1;
  if( bits_.pointer && P0_ != NULL ) P0_[0] = val;
  else { shape_ = val; bits_.pointer = 0; }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromParent::operator="   
PromParent &PromParent::operator=( const PromParent &a )
{
  assert(this!=NULL);
  
  procID_ = a.procID_;
  gIndex_ = a.gIndex_;
  geq_ = a. geq_;                          assert( geq_>=0);
  PromElement::operator=( a ); bits_.shp_sz = 0;

  if( a.bits_.shp_sz > 1 ) {
    if( a.P0_ != NULL ){
      SetPsz( a.bits_.shp_sz );
      assert( a.bits_.shp_sz == bits_.shp_sz );
      assert( a.bits_.pointer && bits_.pointer );
      for( unsigned int ii = 0; ii < bits_.shp_sz ; ii++ ) P0_[ii] = a.P0_[ii];
    }
    else P0_ = NULL;
    if( a.DAP_ != NULL ) {
      PetscMalloc( sizeof(shpfloat)*bits_.shp_sz, &DAP_ ); 
      for( unsigned int ii = 0; ii < bits_.shp_sz ; ii++ ) DAP_[ii] = a.DAP_[ii];
      if( bits_.shp_sz != a.bits_.shp_sz ) {
	assert(bits_.shp_sz == 0);
	SetPsz_hard( a.bits_.shp_sz );
	P0_ = NULL;
      }
    }
    else DAP_ = NULL;
  }
  else if( a.bits_.shp_sz == 1) { 
    setScalarP0( a.bits_.pointer ? a.P0_[0] : a.shape_ ) ; 
  }  
  else if ( a.bits_.shp_sz == 0) {
    SetPsz( 0 );
    if( DAP_ != NULL ) PetscFree(DAP_); DAP_ = NULL;
  }

  return *this;
}

int PromParent::Copy( const PromNode * const node, const int mype )
{
  assert(this!=NULL); int ierr;

  gIndex_ = node->index_.global(ierr); // can be an env node
  geq_ = node->geq(ierr);              // assert( geq_>=0);
  int oldsz = bits_.shp_sz;
  PromElement::operator=( *node );
  bits_.shp_sz = oldsz;
  assert(node->symbData_!=NULL);
  if( node->isGhost() ) setProc(node->symbData_->ghostID.proc());
  else setProc(mype);
  kill(FALSE);
  mark(FALSE);

  return 0;
}

//////////////////////////////////////////////////////////////////////
// PromParentArr
//
#undef __FUNCT__  
#define __FUNCT__ "PromParentArr::PromParentArr"
/* PromParentArr::PromParentArr ***************************************
 *
 */
PromParentArr::PromParentArr( PromParentBank * const bank, const int n ) :
  data_(NULL), np_(0)
{
  if( n > 0 ) SetSize( bank, n );
}

#undef __FUNCT__  
#define __FUNCT__ "PromParentArr::~PromParentArr"
/* PromParentArr::~PromParentArr ***************************************
 *
 */
PromParentArr::~PromParentArr()
{
  if( data_ != NULL ) PetscFree( data_ );
}

#undef __FUNCT__  
#define __FUNCT__ "PromParentArr::SetSize"
/* PromParentArr::SetSize ***************************************
 *
 */
int PromParentArr::SetSize( PromParentBank * const bank, int nn )
{
  assert(this!=NULL);  assert(nn>0 && nn != np_);
  int k, ierr;
  PromParent **pp; 
  ierr = PetscMalloc(nn*sizeof(PromParent*),&pp); CHKERRQ(ierr);
  
  if( nn > np_ ){ // adding
    for( k = 0 ; k < np_ ; k++ ) pp[k] = data_[k]; // copy pointer to old
    for(;k<nn;k++){ bank->GetFreePar( &pp[k] ); assert(pp[k]!=NULL);} // alloc rest
  } 
  else { // nn < np_
    for( k = 0 ; k < nn ; k++ ) pp[k] = data_[k]; // copy (some) old
    while(k++ < np_) bank->Deallocate( data_[k-1] );  // delete rest
  }
  if( data_ != NULL ) PetscFree( data_ );
  data_ = pp;
  np_ = nn;
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromParentArr::GetWrite_size"
/* PromParentArr::GetWrite_size ***************************************
 *
 */
int PromParentArr::GetWrite_size( int &sz )const
{
  sz = 0;         // < *<par [, -4]> >
  for( int ii = 0, tt ; ii < np_ ; ii++ ){
    int ierr = data_[ii]->GetWrite_size(tt);  CHKERRQ(ierr);
    if( tt%2 != 0 ) tt++;
    sz += tt;
  }
  assert(sz%2==0);
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromParentArr::Read"
/* PromParentArr::Read ***************************************
 *
 */
int PromParentArr::Read( int **ppb )
{
  assert(this!=NULL);
  int ierr, ii, sz, *pb = *ppb; 
  assert(data_ != NULL);
  if((size_t)pb%sizeof(shpfloat)!=0)SETERRQ(1,"non-double aligned pointer");

  // < *<par [, -4]> >
  for( ii = 0 ; ii < np_ ; ii++ ){
    ierr = data_[ii]->Read( &pb ); CHKERRQ(ierr);
    if( (size_t)pb%sizeof(shpfloat) != 0 ){ sz = *pb++; assert(sz==-4); }
  }
  *ppb = pb;

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromParentArr::Write"
/* PromParentArr::Write ***************************************
 *
 */
int PromParentArr::Write( int **ppb )const
{
  int ierr, ii, *pb = *ppb; 
  if((size_t)pb%sizeof(shpfloat)!=0)SETERRQ(1,"non-double aligned pointer");
  
  // < *<par, [, -4]> >
  for( ii = 0 ; ii < np_ ; ii++ ){
    ierr = data_[ii]->Write( &pb ); CHKERRQ(ierr);
    if( (size_t)pb%sizeof(shpfloat) != 0 ) *pb++ = -4;
  }
  *ppb = pb;

  return 0;
}

//////////////////////////////////////////////////////////////////////
// PromMPINode
//
#undef __FUNCT__  
#define __FUNCT__ "PromMPINode::Write/Read" 
/* PromMPINode::Write/Read ******************************************
 *
 */
int PromMPINode::Write( int **ppb ) const
{
  int ierr, fsz = sizeof(gfloat)/sizeof(int), *pb0 = *ppb;

  // coord, index, gost.p, bits, geq, findex
  ierr = coord_.Write( ppb );         CHKERRQ(ierr); 
  ierr = index_.Write( ppb );         CHKERRQ(ierr); 
  ierr = ghostID_.Write( ppb );       CHKERRQ(ierr); 
  ierr = PromElement::Write( ppb );   CHKERRQ(ierr);
  *(*ppb)++ =  geq_; *(*ppb)++ = fIndex_; 
  if( (*ppb-pb0)%fsz != 0 ) *(*ppb)++;
  PromContext::Write( angle_, *ppb );

  return 0;
}
int PromMPINode::Read( int **ppb ) 
{
  int ierr, fsz = sizeof(gfloat)/sizeof(int), *pb0 = *ppb;

  // coord, index, gost.p, bits, geq, findex
  ierr = coord_.Read( ppb );        CHKERRQ(ierr); 
  ierr = index_.Read( ppb );        CHKERRQ(ierr); 
  ierr = ghostID_.Read( ppb );       CHKERRQ(ierr); 
  // elemid -- keep some old
  ierr = PromElement::Read( ppb );       CHKERRQ(ierr); 
  geq_ = *(*ppb)++;   fIndex_ = *(*ppb)++; 
  if( (*ppb-pb0)%fsz != 0 ) *(*ppb)++ = -2; 
  PromContext::Read( angle_, *ppb );

  return 0;
} 

int PromMPINode::GetWrite_size( int &sz ) 
{ 
  int ii, jj, kk, cc, ierr, fsz = sizeof(gfloat)/sizeof(int); 

  ierr = PromElement::GetWrite_size(ii);        CHKERRQ(ierr); 
  ierr = PromLocalID::GetWrite_size(jj);      CHKERRQ(ierr); 
  ierr = PromGlobalID::GetWrite_size(kk);       CHKERRQ(ierr); 
  ierr = PromCoord::GetWrite_size(cc);    CHKERRQ(ierr); 
  sz = ii + jj + kk + cc + 2;
  if( sz%fsz != 0 ) sz++;
  sz += fsz; // angle_

  return 0;
}   

////////////////////////////////////////////////////////////////////////
// Archive methods
//
#undef __FUNCT__  
#define __FUNCT__ "PromParent::Archive" 
/* PromParent::Archive() ***********************************************
 *
 *   - Save/Write to file 
 *
 *  INPUT:
 *    -  file: file to read/write
 *    -  type: READ, WRITE, PRINT(pretty debug output)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code.
 */
int PromParent::Archive(FILE *file, const PromArchiveType type/*=PROM_PRINT*/)
{
  PromElement::Archive( file, type );
  if( type == PROM_PRINT ) {
    PetscPrintf(PETSC_COMM_SELF,"PromParent<proc=%d gid=%d>", procID_,gIndex_);
  }
  else if( type == PROM_WRITE ) {
    fprintf( file, "%d %d ", procID_, gIndex_ );
    // write shapes!!!
  }
  else { 
    assert( type == PROM_READ );
    fscanf( file, "%d %d ", &procID_, &gIndex_ );
  }
  
  return 0;
}
#undef __FUNCT__  
#define __FUNCT__ "PromNode::Archive" 
/* PromNode::Archive() ***********************************************
 *
 *   - Save/Write to file 
 *
 *  INPUT:
 *    -  file: file to read/write
 *    -  type: READ, WRITE, PRINT(pretty debug output)
 *
 *  SIDE EFFECTS: 
 *  RETURN:
 *    - PETSc error code.
 */
int PromNode::Archive(FILE *file, const PromArchiveType type/*=PROM_PRINT*/)
{
  int ierr, ii;

  ierr = PromElement::Archive( file, type );                   CHKERRQ(ierr);
  ierr = index_.Archive( file, type );                      CHKERRQ(ierr);
  for( ii = 0 ; ii < nParents() ; ii++ ) {
    ierr = parents_->data_[ii]->Archive( file, type );       CHKERRQ(ierr);
  } 
  if( type == PROM_PRINT ) {
    //  PromSymbData *symbGeoData_
    //fprintf( file,"PromGlobalID<ndf=%d>\n",ndf_);
  }
  else if( type == PROM_WRITE ) {
    //  PromSymbData *symbGeoData_
    //fprintf( file,"ndf=%d ", ndf_ );
  }
  else { 
    //  PromSymbData *symbGeoData_
    assert( type == PROM_READ );
    //fscanf( file, "ndf=%d ", &ndf_ );    
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromAggNode::Archive" 
/* PromAggNode::Archive() ***********************************************
 *
 *   - Save/Write to file 
 *
 *  INPUT:
 *    -  file: file to read/write
 *    -  type: READ, WRITE, PRINT(pretty debug output)
 *
 *  SIDE EFFECTS: 
 *  RETURN:
 *    - PETSc error code.
 */
int PromAggNode::Archive(FILE *file, const PromArchiveType type/*=PROM_PRINT*/)
{
  int ierr;

  ierr = PromNode::Archive( file, type );                      CHKERRQ(ierr);
  if( type == PROM_PRINT ) {
    
  }
  else if( type == PROM_WRITE ) {
    PromContext::printArr( file, B_, Bsz_, "B[%d]");
  }
  else { 
    PromContext::scanArr( file, &B_, &Bsz_, "B[%d]");
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoNode::Archive" 
/* PromGeoNode::Archive() ***********************************************
 *
 *   - Save/Write to file 
 *
 *  INPUT:
 *    -  file: file to read/write
 *    -  type: READ, WRITE, PRINT(pretty debug output)
 *
 *  SIDE EFFECTS: 
 *  RETURN:
 *    - PETSc error code.
 */
int PromGeoNode::Archive(FILE *file, const PromArchiveType type/*=PROM_PRINT*/)
{
  int ierr;

  ierr = PromNode::Archive( file, type );                      CHKERRQ(ierr);
  ierr = symbGeoData_->coord.Archive(file,type);   CHKERRQ(ierr);
  if( type == PROM_PRINT ) {
    
  }
  else if( type == PROM_WRITE ) {
    
  }
  else { 
    
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGlobalID::Archive" 
int PromGlobalID::Archive( FILE *file, 
			   const PromArchiveType type/*=PROM_PRINT*/)
{
  if( type == PROM_PRINT ) {
    fprintf( file,"PromGlobalID<pid=%d,proc=%d>\n", lIndex_, procID_ );
  }
  else if( type == PROM_WRITE ) {
    fprintf( file,"%d %d", lIndex_, procID_ );
  }
  else { 
    assert( type == PROM_READ );
    fscanf( file, "%d %d ", &lIndex_, &procID_ );
  }
  return 0;
}
